
""" Utils for implement and dealing with exceptions. """

from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
  from nr.util.generic import T


def safe_str(func: t.Callable[[T], str]) -> t.Callable[[T], str]:
  """ Decorator for a #__str__() method of an #Exception subclass that catches an exception that occurs in the
  string formatting function, logs it and returns the message of the occurred exception instead. """

  import functools
  import logging

  @functools.wraps(func)
  def wrapper(self) -> str:
    try:
      return func(self)
    except Exception as exc:
      type_name = type(self).__module__ + '.' + type(self).__name__
      logger = logging.getLogger(type_name)
      logger.exception('Unhandled exception in %s.__str__()', type_name)
      return str(exc)

  return wrapper




""" Tools for URL handling. """

from __future__ import annotations

import dataclasses
import urllib.parse


@dataclasses.dataclass
class Url:
  """ Helper to represent the components of a URL, including first class support for username, password, host and port. """

  scheme: str = ''
  hostname: str = ''
  path: str = ''
  params: str = ''
  query: str = ''
  fragment: str = ''

  username: str | None = None
  password: str | None = None
  port: int | None = None

  def __str__(self) -> str:
    return urllib.parse.urlunparse((self.scheme, self.netloc, self.path, self.params, self.query, self.fragment))

  @property
  def netloc(self) -> str:
    """ Returns the entire network location with auth and port. """

    auth = self.auth
    if auth:
      return f'{self.auth}@{self.netloc_no_auth}'

    return self.netloc_no_auth

  @property
  def auth(self) -> str | None:
    """ Returns just the auth part of the network location."""

    if self.username or self.password:
      return f'{urllib.parse.quote(self.username or "")}:{urllib.parse.quote(self.password or "")}'

    return None

  @property
  def netloc_no_auth(self) -> str:
    """ Returns the network location without the auth part."""

    if self.port is None:
      return self.hostname

    return f'{self.hostname}:{self.port}'

  @staticmethod
  def of(url: str) -> Url:
    """ Parses the *url* string into its parts.

    Raises:
      ValueError: If an invalid URL is passed (for example if the port number cannot be parsed to an integer).
    """
    parsed = urllib.parse.urlparse(url)
    return Url(
      scheme=parsed.scheme,
      hostname=parsed.hostname or '',
      path=parsed.path,
      params=parsed.params,
      query=parsed.query,
      fragment=parsed.fragment,
      username=parsed.username,
      password=parsed.password,
      port=parsed.port,
    )




"""
Provides classes that proxy all operations to a delegate object, allowing to implement behaviours at runtime where
the object contained in a "variable" can be interchanged or different based on the async context or thread.
"""

from ._base import BaseProxy, get, get_name
from ._contextlocal import ContextLocalProxy, contextlocal
from ._proxy import Proxy, bind, is_bound, proxy, set_value
from ._stackable import StackableProxy, empty, pop, push
from ._threadlocal import ThreadLocalProxy, threadlocal

__all__ = [
  'get_name', 'get', 'BaseProxy',
  'bind', 'is_bound', 'proxy', 'set_value', 'Proxy',
  'empty', 'push', 'pop', 'StackableProxy',
  'threadlocal', 'ThreadLocalProxy',
  'contextlocal', 'ContextLocalProxy',
]




from __future__ import annotations

import deprecated

from nr.util.generic import T
from nr.util.preconditions import _get_message, _Message, check_not_none


@deprecated.deprecated('use `nr.util.preconditions.check_not_none()` instead')
def assure(v: T | None, msg: _Message | None = None) -> T:
  """ Assures that *v* is not `None` and returns it. If the value is in fact `None`, a {@link ValueError} will
  be raised raised. """

  if v is None:
    raise ValueError(_get_message(msg) if msg else 'expected value to not be None')
  return v




from __future__ import annotations

import abc
import typing as t

import typing_extensions as te

from nr.util.generic import T, T_co, T_contra

if t.TYPE_CHECKING:
  from nr.util.once import Once


T_Comparable = t.TypeVar('T_Comparable', bound='Comparable')
class Comparable(te.Protocol):
  def __lt__(self, other: t.Any) -> bool: ...


T_Consumer = t.TypeVar('T_Consumer', bound='Consumer')
class Consumer(abc.ABC, t.Generic[T]):

  @abc.abstractmethod
  def __call__(self, value: T) -> t.Any: ...

  @staticmethod
  def of(func: t.Callable[[T], t.Any]) -> Consumer[T]:
    return t.cast(Consumer[T], func)


T_Predicate = t.TypeVar('T_Predicate', bound='Predicate')
class Predicate(abc.ABC, t.Generic[T_contra]):

  @abc.abstractmethod
  def __call__(self, obj: T_contra) -> bool: ...

  @staticmethod
  def of(func: t.Callable[[T_contra], bool]) -> Predicate[T_contra]:
    return t.cast(Predicate[T_contra], func)


T_Supplier = t.TypeVar('T_Supplier', bound='Supplier')
class Supplier(abc.ABC, t.Generic[T_co]):

  @abc.abstractmethod
  def __call__(self) -> T_co: ...

  @staticmethod
  def of(func: t.Callable[[], T_co]) -> Supplier[T_co]:
    return t.cast(Supplier[T_co], func)

  @staticmethod
  def once(func: t.Callable[[], T_co]) -> Once[T_co]:
    from nr.util.once import Once
    return Once(Supplier.of(func))




""" General purpose utility library. """

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = "0.8.10"


from .chaindict import ChainDict

# TODO (@NiklasRosenstein): Remove these backwards compatibility imports in a major version bump
from .functional._coalesce import coalesce
from .optional import Optional
from .orderedset import OrderedSet
from .refreshable import Refreshable
from .stream import Stream




__author__ = 'Niklas Rosenstein <nrosenstein@palantir.com>'
__version__ = '0.0.4'

import typing as t

_T = t.TypeVar('_T')
_Message = t.Union[str, t.Callable[[], str]]
_Types = t.Union[type, t.Tuple[type, ...]]


def _get_message(message: _Message) -> str:
  if isinstance(message, str):
    return message
  else:
    return message()


def _repr_types(types: _Types) -> str:
  if isinstance(types, type):
    return types.__name__
  else:
    return '|'.join(t.__name__ for t in types)


def check_argument(value: bool, message: _Message) -> None:
  """
  Raise a #ValueError if the specified *value* is `False`. The specified *message* will be used as
  the message for the produced exception. The *message* may either be a string or an arg-less
  function that produces a string.
  """

  if not value:
    raise ValueError(_get_message(message))


def check_not_none(value: t.Optional[_T], message: _Message = None) -> _T:
  """
  Raises a #ValueError if *value* is `None`.
  """

  if value is None:
    raise ValueError(_get_message(message or 'cannot be None'))
  return value


@t.overload
def check_instance_of(value: t.Any, type_: t.Type[_T], message: _Message = None) -> _T: ...


@t.overload
def check_instance_of(value: t.Any, types: t.Tuple[t.Type, ...], message: _Message = None) -> t.Any: ...


def check_instance_of(value, types, message = None):
  """
  Raises a #TypeError if *value* is not an instance of the specified *types*. If no message is
  provided, it will be auto-generated for the given *types*.
  """

  if not isinstance(value, types):
    if message is None:
      message = f'expected {_repr_types(types)}, got {type(value).__name__} instead'
    raise TypeError(_get_message(message))
  return value


def check_subclass_of(cls: t.Type, types: _Types, message: _Message = None) -> t.Type:
  """
  Raises a #TypeError if *cls* is not a subclass of one of the specified *types*. If *cls* is not
  a type, a different #TypeError is raised that does not include the specified *message*.
  """

  check_instance_of(cls, type)
  if not issubclass(cls, types):
    if message is None:
      message = f'{cls.__name__} is not a subclass of {_repr_types(types)}'
    raise TypeError(_get_message(message))
  return cls




import typing as t

import deprecated
import typing_extensions as te

from nr.util.generic import T_contra


@deprecated.deprecated('use `nr.util.types.Predicate` instead')
class Predicate(te.Protocol[T_contra]):

  def __call__(self, obj: T_contra) -> bool:
    ...
